import logging
from pathlib import Path

import click
from colorlog import ColoredFormatter  # type: ignore

from motllo.build import materialise_structure, process_markdown
from motllo.markdown import (build_markdown, build_tree, full_gitignore,
                             text_tree, write_markdown)

logger = logging.getLogger("motllo")


def configure_logger():
    """Fancy logging is nicer"""
    formatter = ColoredFormatter(
        "%(log_color)s%(levelname)s - %(message)s",
        datefmt=None,
        reset=True,
        log_colors={
            "DEBUG": "yellow",
            "INFO": "cyan",
            "WARNING": "purple",
            "ERROR": "red",
            "CRITICAL": "red,bg_white",
        },
        secondary_log_colors={},
        style="%",
    )
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)


@click.group()
@click.option("--debug", help="Set log level to debug", is_flag=True)
def cli(debug):
    """Motllo generates Markdown files that can be used as repository/folder templates and can also convert them into a full folder/repository structure. Use the `markdown` command to convert any path into a Markdown file, that is then almost completely editable (and human readable!). You can then run the `build` command using that file to recreate the repository anywhere.

    The goal is to have templates that are small, self-contained and literate. In each section of the Markdown document you can intersperse prose with code, and all code blocks will be weaved together, try it out!

    With motllo you can create templates that can be stored together, easy to read and modify on a whim.
    """
    configure_logger()
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)


@click.argument("path")
@click.option(
    "--gitignore/--no-gitignore",
    default=True,
    help="Use local and global gitignores, yes by default",
)
@click.option(
    "--ignore",
    help='Glob patterns to ignore, comma separated between quotes like "*.py,*.c,*.scala"',
)
@click.option(
    "--force-include",
    help='Glob patterns to forcefully include, comma separated between quotes like "*.py,*.c,*.scala"',
)
@click.option("-o", "--output", help="Destination markdown file", required=True)
@click.option(
    "-x",
    "--max-length",
    help="Maximum amount of lines to write in the markdown, for each file. Use -1 for `all of them`. Defaults to 15",
    type=int,
    default=15,
)
@cli.command()
def markdown(path, gitignore, ignore, output, max_length, force_include):
    """Generate a Markdown template from a folder or repository at PATH. Will
ignore hidden files, you can use --force-include to add them"""
    ppath = Path(path)
    opath = Path(output)
    if ignore is not None:
        ignore_globs = ignore.split(",")
    else:
        ignore_globs = []
    if force_include is not None:
        include_globs = force_include.split(",")
    else:
        include_globs = None
    if gitignore:
        gitignore_globs = full_gitignore(ppath) + ["output"]
    else:
        gitignore_globs = []
    all_ignore_globs = gitignore_globs + ignore_globs
    if len(all_ignore_globs) == 0:
        all_ignore_globs = None
    try:
        structure = build_tree(
            Path.cwd() / ppath,
            ignore_globs=all_ignore_globs,
            include_globs=include_globs,
        )
    except Exception as exc:
        logger.error("Uncaught exception building the tree: %s", exc)
    try:
        generated_markdown = build_markdown(structure, max_length)
    except Exception as exc:
        logger.error("Uncaught exception generating the Markdown: %s", exc)
    try:
        write_markdown(generated_markdown, opath)
    except Exception as exc:
        logger.error("Problem writing Markdown file: %s", exc)


@click.argument("path")
@click.option(
    "--dry-run/--commit",
    default=True,
    help="Dry run by default so you can see what it does",
)
@click.option(
    "--ignore-existing-folders",
    is_flag=True,
    default=False,
    help="Ignore if the destination folder already exists",
)
@click.option(
    "-r",
    "--replace",
    help='Multiple replacement rules separated by colons, like -r "$PROJ:world_domination", -r"$TOOLS:python"',
    multiple=True,
)
@click.option(
    "-o", "--output", help="Destination path to create everything", required=True
)
@cli.command()
def build(path, dry_run, output, replace, ignore_existing_folders):
    """Build a file/folder structure based on a Markdown document at PATH"""
    ppath = Path(path)
    opath = Path(output)
    if dry_run:
        logger.info("Dry run materialising %s to %s", ppath, opath)
        logger.info(
            "The following files and folders will be created, according to the replacements you may have specified"
        )
    replacements = {}
    for rule in replace:
        key, value = rule.split(":")
        replacements.update({key: value})
    try:
        structure = process_markdown(ppath, replacements=replacements)
    except Exception as exc:
        logger.exception("Uncaught exception processing Markdown at path: %s", exc)
    try:
        materialise_structure(
            structure,
            opath,
            dry_run=dry_run,
            replacements=replacements,
            ignore_existing_folders=ignore_existing_folders,
        )
        if dry_run:
            logger.warning(
                "If the above looks good, add the flag --commit to create the files and folders"
            )
    except Exception as exc:
        logger.exception("Uncaught exception materialising Markdown at path: %s", exc)


@click.argument("path")
@click.option(
    "--gitignore/--no-gitignore",
    default=True,
    help="Use local and global gitignores, yes by default",
)
@click.option(
    "--ignore",
    help='Glob patterns to ignore, comma separated between quotes like "*.py,*.c,*.scala"',
)
@click.option(
    "--force-include",
    help='Glob patterns to forcefully include, comma separated between quotes like "*.py,*.c,*.scala"',
)
@cli.command()
def tree(path, gitignore, ignore, force_include):
    """Generate only the visual folder tree (like the UNIX tree command)"""
    ppath = Path(path)
    if ignore is not None:
        ignore_globs = ignore.split(",")
    else:
        ignore_globs = []
    if force_include is not None:
        include_globs = force_include.split(",")
    else:
        include_globs = None
    if gitignore:
        gitignore_globs = full_gitignore(ppath) + ["output"]
    else:
        gitignore_globs = []
    all_ignore_globs = gitignore_globs + ignore_globs
    if len(all_ignore_globs) == 0:
        all_ignore_globs = None
    try:
        only_tree = text_tree(
            Path.cwd() / ppath,
            ignore_globs=all_ignore_globs,
            include_globs=include_globs,
        )
    except Exception as exc:
        logger.exception("Uncaught exception generating the tree: %s", exc)
    click.echo(only_tree)


if __name__ == "__main__":
    cli()



from typing import Any, Dict, List, Optional


class Node:
    """Just to make mypy happy"""


class Contents:
    """Simulated context manager for file.open for tests and helpers"""

    def __init__(self, contents=None):
        self.contents = contents

    def read(self):
        """Just returns the contents"""
        return "\n".join(self.contents)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


class File(Node):
    """Simulated File node, mimicking the API from Path needed to test and process a tree"""

    EMPTY_CONTENTS = "This file is empty, mode was {mode}"

    def __init__(
        self, path, basename="", replacements: Optional[List[Dict[str, str]]] = None
    ):
        self._path = ""
        self.name = ""
        self._rename(path)
        self.contents: Optional[List[str]] = None
        self.replacements = replacements
        splitted = self.name.split(".")
        if len(splitted) < 2:
            self.suffix = None
        else:
            self.suffix = splitted[-1]
        self.basename = basename

    def _rename(self, path):
        self._path = path
        self.name = path

    def as_posix(self):
        """Not really posix, but useful to generate links in the markdown result"""
        if self.basename == "":
            return self.name
        return self.basename + "/" + self.name

    def open(self, mode="meh"):
        """Simulates a content open"""
        if self.contents is None:
            return Contents(contents=[self.EMPTY_CONTENTS.format(mode=mode)])
        return Contents(contents=self.contents)

    def set_contents(self, contents):
        """Adds content as lines"""
        self.contents = contents.strip().split("\n")
        return self

    def set_replacements(self, replacements):
        """Adds replacement rules"""
        self.replacements = replacements
        return self

    @staticmethod
    def is_dir():
        """What do you think this is"""
        return False

    def __repr__(self):
        return f"{self.name}({self.contents})"

    def __eq__(self, other):
        return str(self) == str(other)


class Folder(Node):
    """Simulated File node, mimicking the API from Path needed to test and process a tree"""

    def __init__(self, path, contents: List[Any] = None, depth=0, basename=""):
        self._depth = depth
        self._path = ""
        self._basename = basename
        self.basename = ""
        self.name = ""
        self._rename(path)
        if contents is None:
            self._contents = []
        else:
            self._contents = contents
        self.suffix = ""

    def _rename(self, path):
        self._path = path
        self.name = path
        if self._basename == "":
            self.basename = path
        else:
            self.basename = self._basename + "/" + path

    def __repr__(self):
        return self.name + f"[{self.iterdir()}]"

    def as_posix(self):
        "Returns a kind of _as_posix, used to generate basenames and links"
        return self.basename

    def append_to_contents(self, node: Node):
        """Adds files or folders to a folder"""
        self._contents += [node]

    def prune(self):
        """Removes empty folders"""
        for item in self.iterdir():
            if item.is_dir():
                item.prune()
        self._contents = list(
            filter(lambda x: not (x.is_dir() and x.is_empty()), self._contents)
        )
        return self

    def is_empty(self):
        """Obvious"""
        return len(self._contents) == 0

    @staticmethod
    def is_dir():
        """🙄"""
        return True

    def iterdir(self):
        """Returns the contents so we can iterate inside"""
        return self._contents

    def __eq__(self, other):
        return str(self) == str(other)


SPACE = "    "
BRANCH = "│   "
TEE = "├── "
LST = "└── "


def tree(base: Folder, prefix: str = ""):
    """From https://stackoverflow.com/a/59109706 by https://twitter.com/aaronchall
    """
    contents = list(base.iterdir())
    pointers = [TEE] * (len(contents) - 1) + [LST]
    for pointer, path in zip(pointers, contents):
        yield prefix + pointer + path.name
        if path.is_dir():
            extension = BRANCH if pointer == TEE else SPACE
            yield from tree(path, prefix=prefix + extension)


def url(path):
    """Generate a Markdown link for a file/folder"""
    urlable = path.as_posix().replace("/", "").replace(".", "").lower()
    if path.is_dir():
        return f"`{path.name}`"
    return f"[`{path.name}`](#{urlable})"


def tree_links(base: Folder, prefix: str = ""):
    """Generate a table of contents based on the tree itself"""
    contents = list(base.iterdir())
    pointers = ["- "] * len(contents)
    for pointer, path in zip(pointers, contents):
        yield prefix + pointer + url(path)
        if path.is_dir():
            yield from tree_links(path, prefix=prefix + SPACE)



import logging
from pathlib import Path
from typing import List, Optional

from motllo.ops import File, Folder

logger = logging.getLogger("motllo.path_traverser")


def matches_glob(path: Path, globs: List[str]):
    """Checks if the path is matched by a list of globs"""
    if globs is None:
        return False
    for glob in globs:
        if path.match(glob):
            return True
    return False


class Traverser:
    """Traverser of (real or simulated) folder hierarchy. Callable class, configuration is passed to the constructor"""

    def __init__(self, ignore_globs=None, include_globs=None):
        self.ignore_globs = ignore_globs
        self.include_globs = include_globs
        self.initial_path: Optional[Path] = None

    def __call__(self, initial_path) -> Folder:
        self.initial_path = initial_path
        return self._traverser(self.initial_path, depth=0)

    def _handle_dir(self, path: Path, depth: int) -> Folder:
        logger.debug("Found folder: %s", path.name)
        return self._traverser(path, depth)

    def _handle_file(self, path: Path, base_path: Path) -> File:
        logger.debug("Found file: %s", path.name)
        with path.open("r") as data:
            try:
                contents = data.read()
            except Exception as exc:
                msg = f"Could not read file {path}, {exc}"
                logger.error(msg)
                contents = msg
        if self.initial_path is None:
            initial_as_posix = ""
        else:
            initial_as_posix = self.initial_path.as_posix()
        return File(
            path.name, basename=base_path.as_posix().replace(initial_as_posix, ""),
        ).set_contents(contents)

    def _traverser(self, base_path, depth=0):
        tree = []
        first_level = list(base_path.iterdir())
        logger.debug("Here is the first level: %s", first_level)
        for path in first_level:
            if not matches_glob(path, self.include_globs):
                if matches_glob(path, self.ignore_globs):
                    continue
                if path.name.startswith("."):
                    continue
            if path.is_dir():
                tree += [self._handle_dir(path, depth + 1)]
            else:
                tree += [self._handle_file(path, base_path)]
        return Folder(base_path.name, tree, depth)



import logging
from typing import List, Optional

from motllo.ops import File, Folder

logger = logging.getLogger("motllo.parse_tree")


def find_file_depth(line: str):
    """Find how deep in a hierarchy a file is by checking the tree"""
    for idx, item in enumerate(line):
        if item not in ["│", "├", "─", "└"] and not item.isspace():
            return idx // 4 - 1
    return -1


def nodename(line: Optional[str]):
    """Extract the file/folder name out of the tree representation"""
    if line is None:
        raise Exception("A node name can't be None. Please report this as an issue")
    return line[(find_file_depth(line) + 1) * 4 :]


class TreeParser:
    """Converts a text-tree into a tree of folders and files"""

    def __init__(self, tree):
        self._result = []
        self._previous_depth = -1
        self._previous_line: Optional[str] = ""
        self._tree = tree

    def _handle_folder(self):
        logger.debug("Adding a folder for Prev: %s", self._previous_line)
        if len(self._result) > 0:
            basename = self._result[-1].basename
        else:
            basename = ""
        self._result += [Folder(nodename(self._previous_line), basename=basename)]

    def __call__(self):
        self._parse(self._tree)
        if len(self._result) > 1:
            raise Exception(
                f"Some folders have not been folded properly, please report this issue, {self._result}"
            )
        return self._result[0]

    def _fold_folders(self):
        logger.debug("Closing and folding opened folders, %s", self._result)
        last = self._result.pop()
        self._result[len(self._result) - 1].append_to_contents(last)

    def _handle_last_line_of_folder(self, line):
        logger.debug("Adding last line of a folder and closing it, %s", line)
        last = self._result.pop()
        basename = last.basename
        last.append_to_contents(File(nodename(self._previous_line), basename=basename))
        self._result[self._previous_depth - 1].append_to_contents(last)
        self._previous_line = line

    def _close_opened_folders(self, line, _counter):
        counter = _counter
        logger.debug(
            "Closing and folding %s folders with Prev: %s, Cur: %s",
            counter,
            self._previous_line,
            line,
        )
        self._handle_last_line_of_folder(line)
        counter -= 1
        while counter > 0:
            self._fold_folders()
            counter -= 1

    def _add_to_open_folder(self, line, depth):
        logger.debug("Adding previous to folder, rolling line to next")
        if depth > len(self._result) - 1:
            raise Exception(f"Depth larger than result: {depth} | {self._result}")
        basename = self._result[depth].basename
        self._result[depth].append_to_contents(
            File(nodename(self._previous_line), basename=basename)
        )
        self._previous_line = line

    def _parse(self, tree: List[str]):
        """Parses a tree"""
        if len(tree) == 0:
            self._result = None
        for line in tree:
            logger.debug("Result: %s", self._result)
            depth = find_file_depth(line)
            logger.debug("Prev: %s Cur: %s Depth: %s", self._previous_line, line, depth)
            if depth > self._previous_depth:
                self._handle_folder()
                self._previous_line = line
            elif depth < self._previous_depth:
                self._close_opened_folders(line, self._previous_depth - depth)
            else:
                self._add_to_open_folder(line, depth)
            self._previous_depth = depth
        if self._previous_depth > 0:
            self._close_opened_folders(None, self._previous_depth)
        if self._previous_depth == 0 and self._previous_line is not None:
            logger.debug(
                "Closing and folding the final item with %s", self._previous_line
            )
            basename = self._result[self._previous_depth].basename
            self._result[self._previous_depth].append_to_contents(
                File(nodename(self._previous_line), basename=basename)
            )
